/*
 * Sorts standard input using merge sort algorithm
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define ALLOC_CHUNK        100

const char *errMemory  = "%s: error--cannot allocate enough memory\n";
const char *errParsing = "%s: error--failed to parse input\n";

unsigned *buf;

unsigned *mergesort(unsigned *buf, int len);


int main(int argc, char **argv)
{
    int i, len, r;

    /* Read input reallocating the buffer as soon as we run out of space */
    buf = NULL;
    len = 0;
    i   = 0;
    while ( 1 ) {
        if ( i == len ) { /* allocate more space */
            len += ALLOC_CHUNK;
            buf = (unsigned *) realloc( (void *) buf, len * sizeof(*buf) );
            if ( !buf ) {
                fprintf(stderr, errMemory, *argv);
                return 1;
            }
        }
        r = scanf("%u", buf + i++ );
        if ( r == EOF )
            break;
        else if ( r == 0 ) {
            fprintf(stderr, errParsing, *argv);
            return 1;
        }
    }

    len = i - 1; /* forget the buffer size--we don't need it any more */

    /* Sort the buffer */
    if ( !mergesort(buf, len) ) {
        fprintf(stderr, errMemory, *argv);
        return 1;
    }

    /* Print out the result */
    for ( i = 0; i < len; i++ )
        printf("%u\n", *(buf + i));

    /* Free the mallocs! */
    free((void *) buf);

    return 0;
}



unsigned *buffer;
unsigned *sbuffer; /* temporary buffer for writing concatenation result to */

void concatenate(int first, int second, int last);
void _mergesort(int first, int last);

unsigned *mergesort(unsigned *buf, int len)
{
    buffer = buf;
    sbuffer = (unsigned *) malloc( len * sizeof(*buffer) );
    if ( !sbuffer )
        return NULL;

    _mergesort(0, len - 1);

    free(sbuffer);

    return buf;
}

/* The merge sort function proper */
void _mergesort(int first, int last)
{
    int med;

    if ( first == last )
        return;

    med = ( first + last ) / 2;

    /* Sort two subbuffers */
    _mergesort(first, med);
    _mergesort(med+1, last);

    /* Concatenate the two subbuffers */
    concatenate(first, med+1, last);

    return;
}

/*
 * Concatenates two sorted adjacent subbuffers so that the whole buffer becomes
 * sorted.
 *
 * Parameters:
 *      first       offset of first element of left subbuffer;
 *      second      offset of first element of right subbuffer;
 *      last        offset of last element of right subbuffer.
 */
void concatenate(int first, int second, int last)
{
    int i, j, k;

    i = first;  /* offset within left subbufer */
    j = second; /* offset within right subbufer */
    k = first;  /* offset within temporary buffer */
    while ( 1 ) {
        while ( *(buffer + i) <= *(buffer + j) && i < second )
            *(sbuffer + k++) = *(buffer + i++);
        if ( i == second ) {
            /* Don't need to copy out the rest of right subbufer--it is already
               where it should be, at the end of the right subbufer */
            /* Copy in the concatenated-so-far buffer */
            memcpy( buffer + first, sbuffer + first, (k - first) * sizeof(*buffer) );
            break;
        }

        while ( *(buffer + i) >= *(buffer + j) && j <= last )
            *(sbuffer + k++) = *(buffer + j++);
        if ( j == last + 1 ) {
            /* Copy the rest of left subbufer to the end of right subbuffer
             * (memmove because the block being copied may overlap) */
            memmove( buffer + k, buffer + i, (second - i) * sizeof(*buffer) );
            /* Copy in the concatenated-so-far buffer */
            memcpy( buffer + first, sbuffer + first,
                    (k - first) * sizeof(*buffer) );
            break;
        }
    }

    return;
}



/*
 * Local Variables:
 * compile-command: "gcc mergesort.c && cat unsorted.set | ./a.out > sorted.set"
 * End:
 */
