/*
 * Sorts standard input using tree sort algorithm
 */
#include <stdio.h>
#include <stdlib.h>

#define ALLOC_CHUNK        100

const char *errMemory  = "%s: error--cannot allocate enough memory\n";
const char *errParsing = "%s: error--failed to parse input\n";

unsigned *buf;

void treesort(unsigned *buf, int len);


int main(int argc, char **argv)
{
    int i, len, r;

    /* Read input reallocating the buffer as soon as we run out of space */
    buf = NULL;
    len = 0;
    i   = 0;
    while ( 1 ) {
        if ( i == len ) { /* allocate more space */
            len += ALLOC_CHUNK;
            buf = (unsigned *) realloc( (void *) buf, len * sizeof(*buf) );
            if ( !buf ) {
                fprintf(stderr, errMemory, *argv);
                return 1;
            }
        }
        r = scanf("%u", buf + i++ );
        if ( r == EOF )
            break;
        else if ( r == 0 ) {
            fprintf(stderr, errParsing, *argv);
            return 1;
        }
    }

    len = i - 1; /* forget the buffer size--we don't need it any more */

    /* Sort the buffer */
    treesort(buf, len);

    /* Print out the result */
    for ( i = 0; i < len; i++ )
        printf("%u\n", *(buf + i));

    /* Free the mallocs! */
    free( (void *) buf );

    return 0;
}



unsigned *buffer;
#define BUFFER(i)               *(buffer + i)
/* Swaps x and y through intermediate variable t */
#define SWAP(x, y, t)           t = x, x = y, y = t

void reorder_sort_tree(int first, int last);

/* The tree sort function proper */
void treesort(unsigned *buf, int len)
{
    int i;
    unsigned t;

    buffer = buf;
    len--;  /* from now on this is offset of last element */

    /* Build initial sort tree */
    for ( i = len; i >= 0; i-- )
        reorder_sort_tree(i, len);

    /* Sort the buffer */
    while ( len > 0 ) {
        /* Swap first and last elements */
        SWAP ( BUFFER(0), BUFFER(len), t );
        /* Rebuild sort tree for first (len-1) elements */
        reorder_sort_tree(0, --len);
    }

    return;
}


void reorder_sort_tree(int first, int last)
{
    int first_son, swap;
    unsigned a, b, c, t;

    first_son = first + first + 1; /* (first + 1) * 2 - 1 */

    /* Nothing to do if no descendants */
    if ( first_son > last )
        return;

    /* Easy going if only one descendant */
    if ( first_son == last ) {
        if ( BUFFER(first) < BUFFER(first_son) )
            SWAP ( BUFFER(first), BUFFER(first_son), t );
        return;
    }

    /*
     * The root A has two descendants.  If A is less then any of B or C, we
     * select the greater of B and C and swap that with A:
     *
     *        A
     *      /   \
     *     B     C
     */
    a = BUFFER(first);
    b = BUFFER(first_son);
    c = BUFFER(first_son + 1);

    if ( a >= b && a >= c )
        return;

    /* Choose the greater of B and C */
    if ( b > c )
        swap = first_son;
    else /* c >= b */
        swap = first_son + 1;
    /* Swap root A with the chosen descendant */
    SWAP ( BUFFER(first), BUFFER(swap), t );

    /* Go down the tree recursively */
    reorder_sort_tree(swap, last);

    return;
}



/*
 * Local Variables:
 * compile-command: "gcc treesort.c && cat unsorted.set | ./a.out > sorted.set"
 * End:
 */
